try:
    from shapely.geometry import shape, Point
except ImportError:  # pragma: no cover
    shape, Point = None, None


def match(lon, lat, features):
    if not shape:
        return None, None  # pragma no cover
    point = Point(lon, lat)
    mindist, nearest = None, None
    for feature in features:
        polygon = shape(feature['geometry'])
        if polygon.contains(point):
            return feature, 0

        dist = point.distance(polygon)
        if mindist is None or mindist > dist:
            mindist, nearest = dist, feature

    assert mindist is not None
    return nearest, mindist



import re
import itertools
import collections

import attr
from csvw.dsv import UnicodeWriter
from csvw.dsv import reader as base_reader
from clldutils.misc import lazyproperty
from clldutils.apilib import API
from clldutils.attrlib import valid_re, valid_range
from clldutils import jsonlib
from nexus import NexusReader
from newick import loads as newick_loads
from pyglottolog.references import BibFile
from ete3 import Tree
from ete3.parser.newick import NewickError

from pydplace.util import comma_split, semicolon_split, comma_join, semicolon_join, format_float

__all__ = ['Variable', 'Reference', 'Data', 'Society', 'RelatedSociety', 'Dataset',
           'Taxon', 'Phylogeny', 'Repos', 'HRAF']

ID_PATTERN = re.compile('[A-Za-z]+([0-9]+)?')


def reader(*args, **kw):
    for row in base_reader(*args, **kw):
        if isinstance(row, dict) and None in row:
            raise ValueError('too many columns in row {0}'.format(row))
        yield row


@attr.s
class Object(object):
    @classmethod
    def fields(cls):
        return [f.name for f in attr.fields(cls)]

    def astuple(self):
        return attr.astuple(self)


@attr.s
class Variable(Object):
    id = attr.ib(validator=valid_re(ID_PATTERN))
    category = attr.ib(converter=comma_split)
    title = attr.ib()
    definition = attr.ib()

    type = attr.ib(validator=attr.validators.in_(['Continuous', 'Categorical', 'Ordinal']))
    units = attr.ib()
    source = attr.ib()
    changes = attr.ib()
    notes = attr.ib()
    codes = attr.ib(default=attr.Factory(list))

    @classmethod
    def fields(cls):
        return [f.name for f in attr.fields(cls)][:-1]

    def astuple(self):
        return (
            self.id,
            comma_join(self.category),
            self.title,
            self.definition,
            self.type,
            self.units,
            self.source,
            self.changes,
            self.notes,
        )


@attr.s
class Code(Object):
    var_id = attr.ib(validator=valid_re(ID_PATTERN))
    code = attr.ib()
    description = attr.ib(converter=lambda s: s.strip())
    name = attr.ib()


@attr.s
class Reference(object):
    key = attr.ib()
    pages = attr.ib()

    def __str__(self):
        res = self.key
        if self.pages:
            res += ':{0}'.format(self.pages)
        return res

    def format_cldf(self):
        res = self.key
        if self.pages:
            res += '[{}]'.format(self.pages)
        return res

    @classmethod
    def from_string(cls, s):
        if ':' in s:
            k, _, p = s.partition(':')
        elif '[' in s:
            k, _, p = s[:-1].partition('[')
        else:
            k, p = s, ''
        return cls(k.strip(), p.strip())


@attr.s
class Data(Object):
    soc_id = attr.ib()
    sub_case = attr.ib()
    year = attr.ib()
    var_id = attr.ib()
    code = attr.ib()
    comment = attr.ib()
    references = attr.ib(
        converter=lambda s: [Reference.from_string(ss) for ss in semicolon_split(s)])
    source_coded_data = attr.ib()
    admin_comment = attr.ib()

    def astuple(self):
        return (
            self.soc_id,
            self.sub_case,
            self.year,
            self.var_id,
            self.code,
            self.comment,
            semicolon_join('{0}'.format(ref) for ref in self.references),
            self.source_coded_data,
            self.admin_comment,
        )


@attr.s
class ObjectWithSource(object):
    id = attr.ib()
    name = attr.ib()
    year = attr.ib()
    author = attr.ib()
    reference = attr.ib()
    base_dir = attr.ib()
    url = attr.ib()

    @property
    def dir(self):
        return self.base_dir.joinpath(self.id)

    def __str__(self):
        return '{0.name} ({0.id})'.format(self)


@attr.s
class RelatedSociety(object):
    dataset = attr.ib(converter=lambda s: s.strip())
    name = attr.ib(converter=lambda s: s.strip())
    id = attr.ib(converter=lambda s: s.strip())

    def __attrs_post_init__(self):
        if self.dataset == 'WNAI' and self.id.startswith('J'):
            self.id = self.id.replace('J', 'WNAI')

    def __str__(self):
        return '{0.dataset}: {0.name} [{0.id}]'.format(self)

    @classmethod
    def from_string(cls, s):
        match = re.match('([A-Za-z]+):\s*([^\[]+)\[([^\]]+)\]$', s)
        if not match:
            raise ValueError(s)
        return cls(*match.groups())


@attr.s
class RelatedSocieties(Object):
    id = attr.ib()
    related = attr.ib(converter=lambda s: [
        RelatedSociety.from_string(ss) for ss in semicolon_split(s)])

    def astuple(self):
        return (self.id, semicolon_join('{0}'.format(rs) for rs in self.related))


@attr.s
class HRAF(object):
    name = attr.ib()
    id = attr.ib()

    @classmethod
    def fromstring(cls, s):
        if s:
            name, _, id_ = s.strip().partition('(')
            assert id_.endswith(')')
            return cls(name.strip(), id_[:-1])

    def __str__(self):
        return '{0.name} ({0.id})'.format(self)

    @property
    def url(self):
        return 'http://ehrafworldcultures\.yale\.edu/collection\?owc={0}'.format(self.id)


@attr.s
class Society(Object):
    id = attr.ib(validator=valid_re('[A-Za-z][A-Za-z0-9]+'))
    xd_id = attr.ib(validator=valid_re('xd[0-9]+'))
    pref_name_for_society = attr.ib()
    glottocode = attr.ib(validator=valid_re('[a-z0-9]{4}[0-9]{4}$', nullable=True))
    ORIG_name_and_ID_in_this_dataset = attr.ib()
    alt_names_by_society = attr.ib(converter=comma_split)
    main_focal_year = attr.ib()
    HRAF_name_ID = attr.ib(converter=HRAF.fromstring)
    HRAF_link = attr.ib(
        converter=lambda s: s.strip() or None,
        validator=valid_re(
            r'http://ehrafworldcultures\.yale\.edu/collection\?owc=[A-Z0-9]+|in process',
            nullable=True))
    origLat = attr.ib(converter=float)
    origLong = attr.ib(converter=float)
    Lat = attr.ib(converter=float, validator=valid_range(-90, 90))
    Long = attr.ib(converter=float, validator=valid_range(-180, 180))
    Comment = attr.ib()
    glottocode_comment = attr.ib()

    def __str__(self):
        return '{0.pref_name_for_society} ({0.id})'.format(self)

    def astuple(self):
        return (
            self.id,
            self.xd_id,
            self.pref_name_for_society,
            self.glottocode,
            self.ORIG_name_and_ID_in_this_dataset,
            comma_join(self.alt_names_by_society),
            self.main_focal_year,
            '{0}'.format(self.HRAF_name_ID or ''),
            self.HRAF_link,
            format_float(self.origLat),
            format_float(self.origLong),
            format_float(self.Lat),
            format_float(self.Long),
            self.Comment,
            self.glottocode_comment,
        )


@attr.s
class Dataset(ObjectWithSource):
    type = attr.ib(validator=attr.validators.in_(['cultural', 'environmental']))
    description = attr.ib()

    def _path(self, what):
        return self.dir.joinpath('{0}.csv'.format(what))

    def _read_items(self, what, **kw):
        fname = self._path(what)
        return list(reader(fname, **kw)) if fname.exists() else []

    def _write_items(self, what, attr=None, items=None):
        items = items if items is not None else getattr(self, attr or what)
        if items:
            with UnicodeWriter(self._path(what)) as writer:
                writer.writerow(items[0].__class__.fields())
                for item in items:
                    writer.writerow(item.astuple())

    @lazyproperty
    def data(self):
        return [Data(**d) for d in self._read_items('data', dicts=True)]

    @lazyproperty
    def societies(self):
        return [Society(**d) for d in self._read_items('societies', dicts=True)]

    @lazyproperty
    def _society_dict(self):
        return {soc.id: soc for soc in self.societies}

    def society(self, id_):
        return self._society_dict[id_]

    @lazyproperty
    def society_relations(self):
        return [
            RelatedSocieties(**d) for d in self._read_items('societies_mapping', dicts=True)]

    @lazyproperty
    def variables(self):
        codes = {vid: list(c) for vid, c in itertools.groupby(
            sorted(self._read_items('codes', dicts=True), key=lambda c: c['var_id']),
            lambda c: c['var_id'])}
        res = []
        for v in self._read_items('variables', dicts=True):
            v.setdefault('units', '')
            res.append(Variable(codes=[Code(**c) for c in codes.get(v['id'], [])], **v))
        return res

    def write(self):
        self._write_items('societies')
        self._write_items('data')
        self._write_items('variables')
        self._write_items('codes', items=list(
            itertools.chain(*[var.codes for var in self.variables])))
        self._write_items('societies_mapping', attr='society_relations')


@attr.s
class Taxon(object):
    taxon = attr.ib()
    glottocode = attr.ib()
    xd_ids = attr.ib(converter=comma_split)
    soc_ids = attr.ib(converter=comma_split)
    isocode = attr.ib(default=None)
    properties = attr.ib(default=dict)

    @classmethod
    def from_dict(cls, d):
        kw = {'properties': {}}
        fields = set(f.name for f in attr.fields(cls))
        for k, v in d.items():
            if k in fields:
                kw[k] = v
            else:
                kw['properties'][k] = v
        return cls(**kw)


@attr.s
class Phylogeny(ObjectWithSource):
    scaling = attr.ib()
    source_id = attr.ib()

    @property
    def nexus(self):
        return NexusReader(self.trees.as_posix())

    @property
    def newick(self):
        nexus = self.nexus
        nexus.trees.detranslate()
        newick = re.sub(r'\[.*?\]', '', nexus.trees.trees[0])
        try:
            return newick[newick.index('=') + 1:]
        except ValueError:  # pragma: no cover
            return newick

    @property
    def newick_tree(self):
        return newick_loads(self.newick)[0]

    @property
    def is_glottolog(self):
        return self.id.startswith('glottolog_')

    @property
    def trees(self):
        return self.dir.joinpath('summary.trees')

    @property
    def taxa(self):
        return [Taxon.from_dict(d) for d in reader(self.dir.joinpath('taxa.csv'), dicts=True)]


class Repos(API):
    @lazyproperty
    def datasets(self):
        res = [
            Dataset(base_dir=self.path('datasets'), **r) for r in
            reader(self.path('datasets', 'index.csv'), dicts=True)]
        for ds in res:
            if not ds.description:
                readme = ds.base_dir / ds.id / 'README.md'
                if readme.exists():
                    ds.description = readme.read_text(encoding='utf-8')
        return res

    @lazyproperty
    def _dataset_dict(self):
        return {ds.id: ds for ds in self.datasets}

    def dataset(self, id_):
        return self._dataset_dict[id_]

    @property
    def phylogenies(self):
        return [
            Phylogeny(base_dir=self.path('phylogenies'), **r) for r in
            reader(self.path('phylogenies', 'index.csv'), dicts=True)]

    @property
    def societies(self):
        return {
            s.id: s for s in itertools.chain.from_iterable(d.societies for d in self.datasets)
        }

    @property
    def variables(self):
        return {
            v.id: v for v in itertools.chain.from_iterable(d.variables for d in self.datasets)
        }

    @property
    def sources(self):
        return BibFile(self.path('sources.bib'))

    def write(self):
        for ds in self.datasets:
            ds.write()

    def read_csv(self, *comps, **kw):
        return list(reader(self.path(*comps), **kw))

    def read_json(self, *comps):
        return jsonlib.load(self.path(*comps))

    def iter_data(self, datasets=None, variables=None, societies=None):
        for ds in self.datasets:
            if (datasets is None) or (datasets and ds.id in datasets):
                for record in ds.data:
                    if variables and record.var_id not in variables:
                        continue
                    if societies and record.soc_id not in societies:
                        continue
                    yield record

    def check(self):
        glottolog = {l.id: l for l in
                     self.read_csv('csv', 'glottolog.csv', namedtuples=True)}
        msgs = {'error': [], 'warning': []}

        def _msg(type_, msg, obj=None):  # pragma: no cover
            obj = '{0.__class__.__name__} {0.id}: '.format(obj) if obj else ''
            msgs[type_].append('%s:%s%s' % (type_.upper(), obj, msg))

        def error(msg, obj=None):  # pragma: no cover
            _msg('error', msg, obj=obj)

        def warning(msg, obj=None):  # pragma: no cover
            _msg('warning', msg, obj=obj)

        sources = set(e.key for e in self.sources.iterentries())
        socids, xdids, gcs, varids = \
            set(), collections.defaultdict(set), collections.defaultdict(set), {}
        for ds in self.datasets:
            for soc in ds.societies:
                if soc.id in socids:  # pragma: no cover
                    error('duplicate society ID: {0}'.format(soc.id), ds)
                xdids[soc.xd_id].add(soc.glottocode)
                gcs[soc.glottocode].add(soc.xd_id)
                socids.add(soc.id)
                if soc.glottocode not in glottolog:  # pragma: no cover
                    warning('{0} without valid glottocode {0.glottocode}'.format(soc), ds)
                elif glottolog[soc.glottocode].family_name == 'Bookkeeping':  # pragma: no cover
                    warning('{0} mapped to Bookkeeping language: {0.glottocode}'.format(soc), ds)
            # are there duplicate variables?
            for var in ds.variables:
                if var.id in varids:  # pragma: no cover
                    error('duplicate variable ID: {0}'.format(var.id), ds)
                varids[var.id] = [
                    c.code for c in var.codes] if var.type in ['Categorical', 'Ordinal'] else []

            # are there undefined variables?
            undefined = set([r.var_id for r in ds.data if r.var_id not in varids])
            for u in undefined:  # pragma: no cover
                error('undefined variable ID: {0}'.format(u), ds)

            for d in ds.data:
                if d.var_id not in varids:  # pragma: no cover
                    error('undefined variable ID: {0}'.format(d.var_id), ds)
                elif len(varids[d.var_id]) > 1 \
                        and d.code not in varids[d.var_id]:  # pragma: no cover
                    error('undefined code for variable {0} and society {1}:{2}'.format(
                        d.var_id, d.soc_id, d.code), ds)
                for ref in d.references:
                    if ref.key not in sources:
                        error('undefined source key "{0}" referenced in {1}'.format(
                            ref.key, ds.id), ds)

        for xdid, glottocodes in xdids.items():
            if len(glottocodes - {None}) > 1:  # pragma: no cover
                # No xd_id can be linked to more than one Glottocode!
                error('xd_id {0} mapped to multiple glottocodes {1}'.format(xdid, glottocodes))

        for p in self.phylogenies:
            if p.source_id:
                if p.source_id not in sources:  # pragma: no cover
                    error('{0}: invalid source_id {1}'.format(p.id, p.source_id), p)
            taxa = set()
            for taxon in p.taxa:
                taxa.add(taxon.taxon)
                if taxon.glottocode and taxon.glottocode not in glottolog:
                    error('{0}: invalid glottocode {1}'.format(p.id, taxon.glottocode), p)
                for socid in taxon.soc_ids:
                    if socid not in socids:
                        error('{0}: invalid soc_id {1}'.format(p.id, socid), p)
                for xdid in taxon.xd_ids:
                    if xdid not in xdids:
                        error('{0}: invalid xd_id {1}'.format(p.id, xdid), p)

            if not p.nexus:  # pragma: no cover
                error('{0}: unable to load summary.trees'.format(p.id), p)

            try:
                Tree(p.newick, format=1)
            except NewickError as e:  # pragma: no cover
                error('{0}: invalid newick tree from summary.trees: {1}'.format(p.id, e), p)

            if not p.is_glottolog:
                for node in p.newick_tree.walk():
                    if node.name and node.is_leaf and node.name not in taxa:  # pragma: no cover
                        warning('Leaf label missing in taxa.csv: {0}'.format(node.name), obj=p)

        for key in ['warning', 'error']:
            for msg in msgs[key]:
                print(msg)
        return not bool(msgs['error'])



"""
Create SOURCES.md
"""
from clldutils.markup import Table


def run(args):
    md = ['# Sources', '']
    for datatype in ['datasets', 'phylogenies']:
        md.append('\n## {0}\n'.format(datatype.capitalize()))
        t = Table('Name', 'Reference')
        for obj in getattr(args.repos, datatype):
            if not obj.id.startswith('glottolog_') or obj.id == 'glottolog_global':
                t.append([
                    '[{0}]({1}/{2})'.format(obj.name, datatype, obj.id),
                    obj.reference])
        md.append(t.render(condensed=False))
    args.repos.path('SOURCES.md').write_text('\n'.join(md), encoding='utf-8')



import types
import shutil
import pathlib
import functools

from clldutils.text import split_text

__all__ = ['comma_split', 'semicolon_split', 'remove_subdirs', 'comma_join']


comma_split = functools.partial(split_text, separators=',', strip=True, brackets={})
semicolon_split = functools.partial(split_text, separators=';', strip=True, brackets={})


def _join(sep, *args):
    if len(args) == 1 and isinstance(args[0], (list, tuple, set, types.GeneratorType)):
        args = args[0]
    return sep.join(args)


comma_join = functools.partial(_join, ', ')
semicolon_join = functools.partial(_join, '; ')


def format_float(f):
    return format(float(f), '.6f').rstrip('0').rstrip('.')


def remove_subdirs(d, pattern='*'):
    for sd in pathlib.Path(d).glob(pattern):
        if sd.is_dir():
            shutil.rmtree(str(sd))



"""
Recreate glottolog data files from the current version published at http://glottolog.org
"""
import re
import itertools

from csvw.dsv import UnicodeWriter, reader
from ete3 import Tree
from pyglottolog.api import Glottolog
try:
    # Compatibility with pyglottolog < 2.x
    from pyglottolog.languoids import Level
except ImportError:
    Level = None

from pydplace.util import comma_join, remove_subdirs

__all__ = ['update']

NEXUS_TEMPLATE = """#NEXUS
Begin trees;
    tree {} = {}
end;
"""


def reference(title, year):
    return "Hammarström, Harald & Forkel, Robert & Haspelmath, Martin. {0}. " \
           "{1}. Jena: Max Planck Institute for the Science of Human History. " \
           "https://glottolog.org/".format(title, year)


def write_tree(tree, fname, taxa_in_dplace, societies_by_glottocode):
    if not fname.exists():
        fname.mkdir()
    tree.prune([str(n) for n in taxa_in_dplace])

    with fname.joinpath('summary.trees').open('w', encoding="utf-8") as handle:
        handle.write(NEXUS_TEMPLATE.format(
            tree.name if tree.name else 'UNTITLED',
            tree.write(format=1)
        ))

    with UnicodeWriter(fname.joinpath('taxa.csv')) as writer:
        writer.writerow(['taxon', 'glottocode', 'xd_ids', 'soc_ids'])
        for gc in sorted(taxa_in_dplace):
            socs = societies_by_glottocode[gc]
            writer.writerow([
                gc,
                gc,
                comma_join(sorted(set(s.xd_id for s in socs))),
                comma_join(sorted(s.id for s in socs))])
    return tree


def trees(societies_by_glottocode, langs, outdir, year, title):
    label_pattern = re.compile("'[^\[]+\[([a-z0-9]{4}[0-9]{4})[^']*'")

    def rename(n):
        n.name = label_pattern.match(n.name).groups()[0]
        n.length = 1

    glottocodes = set(societies_by_glottocode.keys())
    glottocodes_in_global_tree = set()
    index = {}
    outdir = outdir / 'phylogenies'
    remove_subdirs(outdir, 'glottolog_*')
    languoids = {}
    families = []
    for lang in langs:
        if not lang.lineage:  # a top-level node
            if not lang.category.startswith('Pseudo '):
                families.append(lang)
        languoids[lang.id] = lang

    glob = Tree()
    glob.name = 'glottolog_global'

    for family in sorted(families, key=lambda f: f.name):
        node = family.newick_node(nodes=languoids)
        node.visit(rename)
        taxa_in_tree = set(n.name for n in node.walk())
        taxa_in_dplace = glottocodes.intersection(taxa_in_tree)
        if not taxa_in_dplace:
            continue  # pragma: no cover

        tree = Tree("({0});".format(node.newick), format=3)
        tree.name = 'glottolog_{0}'.format(family.id)
        if family.level.name == 'family':
            tree = write_tree(tree, outdir / tree.name, taxa_in_dplace, societies_by_glottocode)
            glottocodes_in_global_tree = glottocodes_in_global_tree.union(
                set(n.name for n in tree.traverse()))
            index[tree.name] = dict(
                id=tree.name,
                name='{0} ({1})'.format(family.name, title),
                author='{0} ({1})'.format(title, family.name),
                year=year,
                scaling='',
                reference=reference(title, year),
                url='https://glottolog.org/resource/languoid/id/{}'.format(family.id))
        else:
            glottocodes_in_global_tree = glottocodes_in_global_tree.union(taxa_in_tree)
        glob.add_child(tree)

    # global
    write_tree(
        glob,
        outdir / glob.name,
        glottocodes_in_global_tree.intersection(glottocodes),
        societies_by_glottocode)
    index[glob.name] = dict(
        id=glob.name,
        name='Global Classification ({0})'.format(title),
        author=title,
        year=year,
        scaling='',
        reference=reference(title, year),
        url='https://glottolog.org/')

    index_path = outdir / 'index.csv'
    phylos = list(reader(index_path, dicts=True))
    with UnicodeWriter(index_path) as writer:
        header = list(phylos[0].keys())
        writer.writerow(header)
        for phylo in sorted(phylos, key=lambda p: tuple(p.values())):
            if phylo['id'] in index:
                writer.writerow([index[phylo['id']].get(k, '') for k in header])
                del index[phylo['id']]
            else:
                writer.writerow(list(phylo.values()))

        for id_, spec in sorted(index.items()):
            writer.writerow([spec.get(k, '') for k in header])


def languoids(api, langs, outdir):
    _Level = Level or api.languoid_levels
    with UnicodeWriter(outdir / 'csv' / 'glottolog.csv') as writer:
        writer.writerow([
            'id', 'name', 'family_id', 'family_name', 'iso_code', 'language_id', 'macroarea',
            'lineage', 'level'])
        for lang in sorted(langs, key=lambda l: l.id):
            if lang.level == _Level.language:
                lid = lang.id
            elif lang.level == _Level.dialect:
                for _, lid, level in reversed(lang.lineage):
                    if level == _Level.language:
                        break
                else:  # pragma: no cover
                    raise ValueError
            else:
                lid = None
            writer.writerow([
                lang.id,
                lang.name,
                lang.lineage[0][1] if lang.lineage else '',
                lang.lineage[0][0] if lang.lineage else '',
                lang.iso or '',
                lid,
                getattr(lang.macroareas[0], 'value', lang.macroareas[0].name)
                if lang.macroareas else '',
                '/'.join(gc for _, gc, _ in lang.lineage),
                lang.level.name,
            ])


def update(repos, gl_repos, year, title):
    societies_by_glottocode = {
        gc: list(socs) for gc, socs in itertools.groupby(
            sorted(repos.societies.values(), key=lambda s: s.glottocode),
            lambda s: s.glottocode)}
    api = Glottolog(gl_repos)
    langs = list(api.languoids())
    languoids(api, langs, repos.repos)
    trees(societies_by_glottocode, langs, repos.repos, year, title)



"""
# - iterate over datasets
#   - group in society sets and others
#   - aggregate data grouped by society set
# - write out package per society set
"""


def run(args):
    raise NotImplementedError()



"""
Augment sources.bib with lgcode for use as Glottolog reference provider.
"""
import collections


def run(args):
    refs = collections.defaultdict(set)

    socs = {}
    for id_, soc in args.repos.societies.items():
        socs[id_] = soc

    for d in args.repos.iter_data():
        for ref in d.references:
            soc = socs[d.soc_id]
            refs[ref.key].add((soc.pref_name_for_society, soc.glottocode))

    def add_lgcode(e):
        if e.key in refs:
            e.fields['lgcode'] = ', '.join('{0} [{1}]'.format(*soc) for soc in sorted(refs[e.key]))

    args.repos.sources.visit(add_lgcode)



"""
Assign socities to TDWG regions
"""
import math

from clldutils.jsonlib import update_ordered

from pydplace import geo

try:
    import fiona
except ImportError:  # pragma: no cover
    fiona = None


def run(args):
    if fiona is None:
        raise SystemExit('fiona and shapely must be installed for this command')

    with fiona.collection(str(args.repos.path("geo", "level2-shape/level2.shp")), "r") as source:
        regions = [f for f in source]

    with update_ordered(args.repos.path("geo", "societies_tdwg.json"), indent=4) as soc_tdwg:
        for ds in args.repos.datasets:
            for soc in ds.societies:
                spec = soc_tdwg.get(
                    soc.id, dict(lat=soc.Lat, lon=soc.Long, name=None, code=None))
                if math.isclose(spec['lat'], soc.Lat) \
                        and math.isclose(spec['lon'], soc.Long) \
                        and spec['code']:
                    continue

                region, dist = geo.match(spec['lon'], spec['lat'], regions)
                spec['name'] = region['properties']['REGION_NAM']
                spec['code'] = str(region['properties']['TDWG_CODE'])

                if dist == 0:  # pragma: no cover
                    args.log.info('{0} contained in region {1}'.format(soc, spec['name']))
                else:
                    args.log.warning(
                        'assigning {0} to nearest region {1}, distance {2}'.format(
                            soc, region['properties']['REGION_NAM'], dist))

                soc_tdwg[soc.id] = spec



"""
List datasets
"""
from clldutils.clilib import Table, add_format


def register(parser):
    add_format(parser, default='simple')


def run(args):
    with Table(args, 'id', 'name', 'type', 'variables', 'societies') as t:
        for ds in args.repos.datasets:
            t.append([ds.id, ds.name, ds.type, len(ds.variables), len(ds.societies)])



"""
Update data derived from Glottolog
"""
import datetime
import contextlib

from clldutils.clilib import PathType
from cldfcatalog import Catalog

from pydplace import glottolog


def register(parser):
    parser.add_argument('glottolog', help='clone of glottolog/glottolog', type=PathType(type='dir'))
    parser.add_argument('--glottolog-version', default=None)


def run(args):
    with contextlib.ExitStack() as stack:
        if args.glottolog_version:  # pragma: no cover
            stack.enter_context(Catalog(args.glottolog, tag=args.glottolog_version))
        glottolog.update(
            args.repos,
            args.glottolog,
            str(datetime.date.today().year),
            "Glottolog {0}".format(args.glottolog_version))



from setuptools import setup, find_packages


setup(
    name='pydplace',
    version='2.4.0',
    license='Apache 2.0',
    description='programmatic access to D-PLACE/dplace-data',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    author='Robert Forkel',
    author_email='forkel@shh.mpg.de',
    url='https://d-place.org',
    keywords='data',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    python_requires='>=3.5',
    install_requires=[
        'attrs>=19.1',
        'clldutils>=3.5.0',
        'cldfcatalog',
        'csvw>=1.6',
        'pyglottolog>=3.0',
        'python-nexus>=1.63',
        'pycldf>=1.14',
        'ete3>=3.0.0b34',
        #'pygdal>=1.11.3.3',
        #'fiona',
        #'shapely',
    ],
    extras_require={
        'dev': ['flake8', 'wheel', 'twine'],
        'test': [
            'mock',
            'pytest>=5',
            'pytest-mock',
            'pytest-cov',
            'coverage>=4.2',
        ],
    },
    entry_points={
        'console_scripts': [
            'dplace=pydplace.__main__:main',
        ]
    },
)



"""
Serialize the core D-PLACE data (without the phylogenies) as CLDF StructureDataset.
"""
import sys
import shutil
import collections

from clldutils.clilib import PathType
from cldfcatalog import Catalog, Repository
from pycldf import StructureDataset, Sources


def register(parser):
    parser.add_argument(
        'glottolog',
        metavar='GLOTTOLOG',
        help="clone of glottolog/glottolog",
        type=PathType(type='dir'),
    )
    parser.add_argument(
        'glottolog_version',
        help="tag to checkout glottolog/glottolog to",
    )
    parser.add_argument(
        '--cldf-repos',
        help="clone of d-place/dplace-cldf",
        default='../dplace-cldf',
        type=PathType(type='dir'))
    parser.add_argument('--dev', action='store_true', default=False)
    parser.add_argument('--fix-code-id', action='store_true', default=False)


def run(args):
    cldf = StructureDataset.in_dir(args.cldf_repos / 'cldf')
    if args.glottolog_version != 'test':  # pragma: no cover
        with Catalog(args.glottolog, args.glottolog_version) as glottolog:
            write_metadata(cldf, args, glottolog)
    else:
        write_metadata(cldf, args, None)
    write_schema(cldf)
    cldf.write(**get_data(cldf, args))
    shutil.copy(str(args.repos.path('LICENSE.txt')), str(args.cldf_repos))
    if not args.dev:
        cldf.validate(log=args.log)


def write_metadata(cldf, args, glottolog):
    cldf.properties['dc:bibliographicCitation'] = \
        'Kathryn R. Kirby, Russell D. Gray, Simon J. Greenhill, Fiona M. Jordan, ' \
        'Stephanie Gomes-Ng, Hans-Jörg Bibiko, Damián E. Blasi, Carlos A. Botero, ' \
        'Claire Bowern, Carol R. Ember, Dan Leehr, Bobbi S. Low, Joe McCarter, ' \
        'William Divale, and Michael C. Gavin. (2016). ' \
        'D-PLACE: A Global Database of Cultural, Linguistic and Environmental Diversity. ' \
        'PLoS ONE, 11(7): e0158391. doi:10.1371/journal.pone.0158391.'
    cldf.properties['dc:title'] = 'CLDF Dataset derived from D-PLACE'
    cldf.properties['dc:description'] = \
        'This dataset contains the data from D-PLACE, ' \
        'the Database of Places, Language, Culture and Environment, ' \
        'serialized as CLDF StructureDataset. ' \
        'D-PLACE societies are formally treated as languages, i.e. society metadata is ' \
        'written to the CLDF LanguageTable and the data on language phylogenies ' \
        'aggregated in D-PLACE is excluded.'
    cldf.properties['dc:related'] = 'https://d-place.org'
    cldf.properties['rdf:type'] = 'http://www.w3.org/ns/dcat#Distribution'
    if glottolog:  # pragma: no cover
        cldf.properties['dcat:accessURL'] = Repository(args.cldf_repos).url
        cldf.add_provenance(wasDerivedFrom=[
            Repository(args.repos.repos).json_ld(),
            glottolog.json_ld(),
        ])
    cldf.add_provenance(wasGeneratedBy=[
        collections.OrderedDict([
            ('dc:title', "python"),
            ('dc:description', sys.version.split()[0])])])


def write_schema(cldf):
    cldf.add_component(
        'LanguageTable',
        'xd_id',
        'ORIG_name_and_ID_in_this_dataset',
        {
            "name": "alt_names_by_society",
            "separator": ";"
        },
        'Dataset_ID',
        'main_focal_year',
        'HRAF_Name'
        'HRAF_ID',
        'HRAF_Link',
        'origLat',
        'origLong',
        'Comment',
        'Glottocode_Comment',
    )
    cldf.add_component(
        'ParameterTable',
        'Dataset_ID',
        {
            "name": "Category",
            "separator": ","
        },
        {
            "name": "Type",
            "datatype": {"base": "string", "format": "|".join(
                ['Continuous', 'Categorical', 'Ordinal'])},
        },
        "Units",
        "Source",
        "Changes",
        "Notes",
    )
    cldf.add_component('CodeTable')
    cldf.add_columns(
        'ValueTable',
        'Sub_Case',
        'Year',
        'Source_Coded_Data',
        'Admin_Comment',
        'Dataset_ID',
    )
    cldf['ValueTable', 'value'].null = ['NA']
    cldf.add_table(
        'datasets.csv',
        {'name': 'ID', 'propertyUrl': 'http://cldf.clld.org/v1.0/terms.rdf#id'},
        {'name': 'Name', 'propertyUrl': 'http://cldf.clld.org/v1.0/terms.rdf#name'},
        {'name': 'Description', 'propertyUrl': 'http://cldf.clld.org/v1.0/terms.rdf#description'},
        {'name': 'Type', 'datatype': {'base': 'string', 'format': 'environmental|cultural'}},
        {'name': 'Year', 'propertyUrl': 'http://purl.org/dc/terms/date'},
        {'name': 'Author', 'propertyUrl': 'http://purl.org/dc/elements/1.1/creator'},
        {'name': 'Reference', 'propertyUrl': 'http://purl.org/dc/terms/bibliographicCitation'},
        'URL',
    )
    cldf.add_table(
        'society_relations.csv',
        {'name': 'ID', 'propertyUrl': 'http://cldf.clld.org/v1.0/terms.rdf#id'},
        'Society_ID',
        'Related_Society_ID',
        'Related_Society_External_ID',
        'Related_Society_Name',
        'Related_Society_Dataset',
    )
    cldf.add_foreign_key('society_relations.csv', 'Society_ID', 'LanguageTable', 'ID')
    cldf.add_foreign_key('society_relations.csv', 'Related_Society_ID', 'LanguageTable', 'ID')
    cldf.add_foreign_key('ParameterTable', 'Dataset_ID', 'datasets.csv', 'ID')
    cldf.add_foreign_key('LanguageTable', 'Dataset_ID', 'datasets.csv', 'ID')
    cldf.add_foreign_key('ValueTable', 'Dataset_ID', 'datasets.csv', 'ID')


def get_data(cldf, args):
    relscount = 0
    cldf.sources = Sources.from_file(args.repos.path('sources.bib'))
    categorical_variables = set()
    data = collections.defaultdict(list)
    dsids = [ds.id for ds in args.repos.datasets]
    for ds in args.repos.datasets:
        data['datasets.csv'].append({
            'ID': ds.id,
            'Name': ds.name,
            'Description': ds.description,
            'Type': ds.type,
            'Year': ds.year,
            'Author': ds.author,
            'Reference': ds.reference,
            'URL': ds.url,
        })
        for soc in ds.societies:
            data['LanguageTable'].append({
                'ID': soc.id,
                'Dataset_ID': ds.id,
                'Name': soc.pref_name_for_society,
                'Glottocode': soc.glottocode,
                'Latitude': soc.Lat,
                'Longitude': soc.Long,
                'Comment': soc.Comment,
                'Glottocode_Comment': soc.glottocode_comment,
                'xd_id': soc.xd_id,
                'ORIG_name_and_ID_in_this_dataset': soc.ORIG_name_and_ID_in_this_dataset,
                'alt_names_by_society': soc.alt_names_by_society,
                'main_focal_year': soc.main_focal_year,
                'HRAF_ID': soc.HRAF_name_ID.id if soc.HRAF_name_ID else None,
                'HRAF_Name': soc.HRAF_name_ID.name if soc.HRAF_name_ID else None,
                'HRAF_Link': soc.HRAF_link,
                'origLat': soc.origLat,
                'origLong': soc.origLong,
            })
        for soc in ds.society_relations:
            for rel in soc.related:
                relscount += 1
                data['society_relations.csv'].append({
                    'ID': str(relscount),
                    'Society_ID': soc.id,
                    'Related_Society_ID': rel.id if rel.dataset in dsids else None,
                    'Related_Society_External_ID': rel.id if rel.dataset not in dsids else None,
                    'Related_Society_Name': rel.name,
                    'Related_Society_Dataset': rel.dataset,
                })
        for param in ds.variables:
            data['ParameterTable'].append({
                'ID': param.id.replace('.', '_'),
                'Dataset_ID': ds.id,
                'Name': param.title,
                'Description': param.definition,
                "Category": param.category,
                "Type": param.type,
                "Units": param.units,
                "Source": param.source,
                "Changes": param.changes,
                "Notes": param.notes,
            })
            for code in param.codes:
                if code.code == 'NA':
                    continue
                categorical_variables.add(code.var_id)
                data['CodeTable'].append({
                    'ID': '{}-{}'.format(code.var_id, code.code).replace('.', '_'),
                    'Parameter_ID': code.var_id.replace('.', '_'),
                    'Name': code.name,
                    'Description': code.description,
                })

        codes = set(c['ID'] for c in data['CodeTable'])
        for i, d in enumerate(ds.data, start=1):
            code_id = None \
                if (d.var_id not in categorical_variables) or d.code == 'NA' \
                else '{}-{}'.format(d.var_id, d.code).replace('.', '_')
            if code_id and (code_id not in codes) and args.fix_code_id:
                code_id = None

            data['ValueTable'].append({
                'ID': '{}-{}'.format(ds.id, i),
                'Language_ID': d.soc_id,
                'Parameter_ID': d.var_id.replace('.', '_'),
                'Dataset_ID': ds.id,
                'Code_ID': code_id,
                'Value': d.code,
                'Comment': d.comment,
                'Sub_Case': d.sub_case,
                'Year': d.year,
                'Source': [ref.format_cldf() for ref in d.references],
                'Source_Coded_Data': d.source_coded_data,
                'Admin_Comment': d.admin_comment,
            })
    return data



"""
Extracts subsets of data for further processing.

To filter societies:

> dplace %(prog)s --society Cj4,Cj5,Cj6 output.csv

To filter societies on a given tree:

> dplace %(prog)s --tree gray_et_al2009 output.csv

To filter societies only from a given dataset:

> dplace %(prog)s --dataset EA output.csv
"""
from csvw.dsv import UnicodeWriter


def register(parser):
    parser.add_argument('filename', help='filename', default=None)
    parser.add_argument('--society', help='restrict to these society ids (x,y,z)', default=None)
    parser.add_argument('--tree', help='restrict to this tree', default=None)
    parser.add_argument('--dataset', help='restrict to these datasets (x,y,z)', default=None)
    parser.add_argument('--variable', help='restrict to thes dataset (x,y,z)', default=None)


def run(args):
    datasets = args.dataset.split(",") if args.dataset else None
    variables = args.variable.split(",") if args.variable else None
    societies = args.society.split(",") if args.society else None

    if args.tree:
        for tree in args.repos.phylogenies:
            if args.tree == tree.id:
                break
        else:
            raise SystemExit("Failed to find Tree %s" % args.tree)
        societies = [s for sublist in [t.soc_ids for t in tree.taxa] for s in sublist]

    with UnicodeWriter(f=args.filename) as out:
        header = [
            'ID',
            'XD_ID',
            'Glottocode',
            'Name',
            'OriginalName',
            'FocalYear',
            'Latitude',
            'Longitude',
            'Variable',
            'Value'
        ]
        out.writerow(header)

        socs = args.repos.societies
        for record in args.repos.iter_data(
                datasets=datasets, variables=variables, societies=societies):
            s = socs.get(record.soc_id, None)
            row = [
                s.id,
                s.xd_id,
                s.glottocode,
                s.pref_name_for_society,
                s.ORIG_name_and_ID_in_this_dataset,
                s.main_focal_year,
                s.Lat,
                s.Long,
                record.var_id,
                record.code
            ]
            out.writerow(row)



import sys
import pathlib
import contextlib

from clldutils.loglib import Logging
from clldutils.clilib import get_parser_and_subparsers, register_subcommands, PathType

import pydplace
from pydplace.api import Repos
import pydplace.commands


def main(args=None, catch_all=False, parsed_args=None, log=None):
    parser, subparsers = get_parser_and_subparsers('dplace')
    parser.add_argument(
        '--repos',
        type=PathType(type='dir'),
        default=pathlib.Path('dplace-data'),
        help='Location of clone of D_PLACE/dplace-data')
    register_subcommands(subparsers, pydplace.commands)

    args = parsed_args or parser.parse_args(args=args)

    if not hasattr(args, "main"):
        parser.print_help()
        return 1

    with contextlib.ExitStack() as stack:
        if not log:  # pragma: no cover
            stack.enter_context(Logging(args.log, level=args.log_level))
        else:
            args.log = log
        args.repos = Repos(args.repos)
        try:
            return args.main(args) or 0
        except KeyboardInterrupt:  # pragma: no cover
            return 0
        except Exception as e:  # pragma: no cover
            if catch_all:
                print(e)
                return 1
            raise


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main() or 0)
